/****************************************
 * Sebastien Leriche, Antoine Jacquet   *
 * Module de gestion memoire            *
 * Niveau 1                             *
 ****************************************/

#include <stdlib.h>
#include <string.h>
#include "memoire.h"

int taille_memoire;
int cpt_malloc=0, cpt_free=0;
FILE *fhistorique=NULL;

int initialiser(struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Initialise les differentes listes
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_libre *p_liste_libre;

  fprintf(fhistorique,"  initialiser\n");
  p_liste_libre=malloc(sizeof(struct liste_libre));
  if (p_liste_libre==NULL)
  {
    fprintf(fhistorique,"  initialiser : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  p_liste_libre->bloc.adresse=0;
  p_liste_libre->bloc.taille=taille_memoire;
  p_liste_libre->p_suivant=NULL;
  *pp_liste_libre=p_liste_libre;
  *pp_liste_tas=NULL;
  return(0);
}

int terminer(struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Libere correctement la memoire
   Retourne -1 en cas de mauvaise gestion de la memoire, 0 sinon */
{
  struct liste_libre *bus_libre=*pp_liste_libre;
  struct liste_tas *bus_tas=*pp_liste_tas;

  fprintf(fhistorique,"  terminer\n");
  while (bus_libre!=NULL)
  {
    *pp_liste_libre=(*pp_liste_libre)->p_suivant;
    free(bus_libre);
    cpt_free++;
    bus_libre=*pp_liste_libre;
  }
  while (bus_tas!=NULL)
  {
    *pp_liste_tas=(*pp_liste_tas)->p_suivant;
    free(bus_tas);
    cpt_free++;
    bus_tas=*pp_liste_tas;
  }
  fprintf(fhistorique,"  terminer : cpt_malloc=%d, cpt_free=%d\n",cpt_malloc,cpt_free);
  if (cpt_malloc==cpt_free)
  {
    fprintf(fhistorique,"  terminer : bonne gestion de la memoire\n");
    return(0);
  }
  else
  {
    fprintf(fhistorique,"  terminer : mauvaise gestion de la memoire !\n");
    return(-1);
  }
}

int allouer(int t, struct liste_libre **pp_liste_libre)
/* Alloue une zone memoire de taille t
   Retourne son adresse, -1 sinon */
{
  struct liste_libre *bus=*pp_liste_libre, *cel_precedente=NULL;
  int adresse;

  fprintf(fhistorique,"    allouer %d\n",t);
  while ((bus!=NULL) && (bus->bloc.taille<t))
  {
    cel_precedente=bus;
    bus=bus->p_suivant;
  }
  if (bus==NULL)
  {
    fprintf(fhistorique,"    allouer : pas de bloc assez grand !\n");
    return(-1);
  }
  adresse=bus->bloc.adresse;
  if (bus->bloc.taille==t)
  {
    if (cel_precedente==NULL) *pp_liste_libre=bus->p_suivant;
    else cel_precedente->p_suivant=bus->p_suivant;
    free(bus);
    cpt_free++;
  }
  else
  {
    bus->bloc.taille-=t;
    bus->bloc.adresse+=t;
  }
  fprintf(fhistorique,"    allouer : adresse=%d\n",adresse);
  return(adresse);
}

int liberer(int adresse_libre, int taille_libre, struct liste_libre **pp_liste_libre)
/* Libere la zone memoire de taille taille_libre a l'adresse adresse_libre
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_libre *cel_precedente=NULL, *cel_suivante=NULL, *bus=*pp_liste_libre, *aux;

  fprintf(fhistorique,"    liberer %d %d\n",adresse_libre,taille_libre);
  if ((adresse_libre<0) || (adresse_libre+taille_libre>taille_memoire))
  {
    fprintf(fhistorique,"    liberer : intervalle invalide ?!\n");
    return(-1);
  }

  /* Recherche des cellules entourant la zone a liberer */
  while ((bus!=NULL) && (bus->bloc.adresse<adresse_libre))
  {
    cel_precedente=bus;
    bus=bus->p_suivant;
  }
  cel_suivante=bus;

  if ((cel_precedente!=NULL) && (adresse_libre<cel_precedente->bloc.adresse+cel_precedente->bloc.taille))
  {
    fprintf(fhistorique,"    liberer : zone deja libre ?!\n");
    return(-1);
  }
  if ((cel_suivante!=NULL) && (adresse_libre+taille_libre>cel_suivante->bloc.adresse))
  {
    fprintf(fhistorique,"    liberer : zone deja libre ?!\n");
    return(-1);
  }

  if ((cel_precedente!=NULL) && (cel_suivante!=NULL) && (cel_suivante->bloc.adresse==cel_precedente->bloc.adresse+cel_precedente->bloc.taille+taille_libre))
  {
    cel_precedente->p_suivant=cel_suivante->p_suivant;
    cel_precedente->bloc.taille+=taille_libre+cel_suivante->bloc.taille;
    free(cel_suivante);
    cpt_free++;
    fprintf(fhistorique,"    liberer : fusion de 3 zones\n");
    return(0);
  }

  if ((cel_precedente!=NULL) && (adresse_libre==cel_precedente->bloc.adresse+cel_precedente->bloc.taille))
  {
    cel_precedente->bloc.taille+=taille_libre;
    fprintf(fhistorique,"    liberer : fusion avec la cellule precedente\n");
    return(0);
  }

  if ((cel_suivante!=NULL) && (cel_suivante->bloc.adresse==adresse_libre+taille_libre))
  {
    cel_suivante->bloc.adresse-=taille_libre;
    cel_suivante->bloc.taille+=taille_libre;
    fprintf(fhistorique,"    liberer : fusion avec la cellule suivante\n");
    return(0);
  }

  aux=malloc(sizeof(struct liste_libre));
  if (aux==NULL)
  {
    fprintf(fhistorique,"    liberer : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  aux->bloc.adresse=adresse_libre;
  aux->bloc.taille=taille_libre;
  aux->p_suivant=cel_suivante;
  if (cel_precedente==NULL) *pp_liste_libre=aux;
  else cel_precedente->p_suivant=aux;
  fprintf(fhistorique,"    liberer : nouvelle cellule inseree\n");
  return(0);
}

int allouer_tas(char identite_tas[], int taille_tas, struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Ajoute un tas de nom identite_tas et de taille taille_tas
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_tas *aux, *bus=*pp_liste_tas;

  fprintf(fhistorique,"  allouer_tas %s %d\n",identite_tas,taille_tas);
  while (bus!=NULL)
  {
    if (strcmp(bus->tas.ident,identite_tas)==0)
    {
      fprintf(fhistorique,"  allouer_tas : tas deja existant !\n");
      return(-1);
    }
    bus=bus->p_suivant;
  }
  aux=malloc(sizeof(struct liste_tas));
  if (aux==NULL)
  {
    fprintf(fhistorique,"  allouer_tas : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  aux->tas.taille=taille_tas;
  strcpy(aux->tas.ident,identite_tas);
  if ((aux->tas.adresse=allouer(taille_tas,pp_liste_libre))==-1)
  {
    free(aux);
    cpt_free++;
    fprintf(fhistorique,"  allouer_tas : erreur de allouer !\n");
    return(-1);
  }
  aux->p_suivant=*pp_liste_tas;
  *pp_liste_tas=aux;
  return(aux->tas.adresse);
}

int liberer_tas(char identite_tas[], struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Libere le tas de nom identite_tas
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_tas *bus=*pp_liste_tas, *precedent=NULL;

  fprintf(fhistorique,"  liberer_tas %s\n",identite_tas);
  while ((bus!=NULL) && (strcmp(identite_tas,bus->tas.ident)!=0))
  {
    precedent=bus;
    bus=bus->p_suivant;
  }
  if (bus==NULL)
  {
    fprintf(fhistorique,"  liberer_tas : tas inexistant !\n");
    return(-1);
  }
  if (liberer(bus->tas.adresse,bus->tas.taille,pp_liste_libre)!=0)
  {
    fprintf(fhistorique,"  liberer_tas : erreur de liberer\n");
    return(-1);
  }
  if (precedent==NULL) *pp_liste_tas=bus->p_suivant;
  else precedent->p_suivant=bus->p_suivant;
  free(bus);
  cpt_free++;
  return(0);
}
