/****************************************
 * Sebastien Leriche, Antoine Jacquet   *
 * Module de gestion memoire            *
 * Niveau 2                             *
 ****************************************/

#include <stdlib.h>
#include <string.h>
#include "memoire.h"

int taille_memoire;
int cpt_malloc=0, cpt_free=0;
FILE *fhistorique=NULL;

struct liste_tas *existe_tas(char ident[], struct liste_tas *bus)
/* Renvoie un pointeur sur le tas de nom ident dans la liste des tas pointee par bus ou NULL sinon */
{
  while (bus!=NULL)
  {
    if (strcmp(bus->tas.ident,ident)==0) return(bus);
    bus=bus->p_suivant;
  }
  return(NULL);
}

int initialiser(struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas, struct programme **pp_liste_programmes)
/* Initialise les differentes listes
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_libre *p_liste_libre;

  fprintf(fhistorique,"  initialiser\n");
  p_liste_libre=malloc(sizeof(struct liste_libre));
  if (p_liste_libre==NULL)
  {
    fprintf(fhistorique,"  initialiser : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  p_liste_libre->bloc.adresse=0;
  p_liste_libre->bloc.taille=taille_memoire;
  p_liste_libre->p_suivant=NULL;
  *pp_liste_libre=p_liste_libre;
  *pp_liste_tas=NULL;
  *pp_liste_programmes=NULL;
  return(0);
}

int terminer(struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas, struct programme **pp_liste_programmes)
/* Libere correctement la memoire
   Retourne -1 en cas de mauvaise gestion de la memoire, 0 sinon */
{
  struct liste_libre *bus_libre=*pp_liste_libre;
  struct liste_tas *bus_tas=*pp_liste_tas;
  struct programme *bus_prog=*pp_liste_programmes;

  fprintf(fhistorique,"  terminer\n");
  while (bus_libre!=NULL)
  {
    *pp_liste_libre=(*pp_liste_libre)->p_suivant;
    free(bus_libre);
    cpt_free++;
    bus_libre=*pp_liste_libre;
  }
  while (bus_tas!=NULL)
  {
    *pp_liste_tas=(*pp_liste_tas)->p_suivant;
    free(bus_tas);
    cpt_free++;
    bus_tas=*pp_liste_tas;
  }
  while (bus_prog!=NULL)
  {
    *pp_liste_programmes=(*pp_liste_programmes)->p_suivant;
    free(bus_prog);
    cpt_free++;
    bus_prog=*pp_liste_programmes;
  }

  fprintf(fhistorique,"  terminer : cpt_malloc=%d, cpt_free=%d\n",cpt_malloc,cpt_free);
  if (cpt_malloc==cpt_free)
  {
    fprintf(fhistorique,"  terminer : bonne gestion de la memoire\n");
    return(0);
  }
  else
  {
    fprintf(fhistorique,"  terminer : mauvaise gestion de la memoire !\n");
    return(-1);
  }
}

int allouer(int t, struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Alloue une zone memoire de taille t
   Retourne son adresse, -1 sinon */
{
  struct liste_libre *bus=*pp_liste_libre, *cel_precedente=NULL;
  int adresse;

  fprintf(fhistorique,"    allouer %d\n",t);

  /* On essaie de trouver un bloc, en appelant le ramasse-miettes si necessaire */
  while ((bus!=NULL) && (bus->bloc.taille<t))
  {
    cel_precedente=bus;
    bus=bus->p_suivant;
  }
  if (bus==NULL)
  {
    ramasse_miettes_compteurs(pp_liste_libre, pp_liste_tas);
    cel_precedente=NULL;
    bus=*pp_liste_libre;
    while ((bus!=NULL) && (bus->bloc.taille<t))
    {
      cel_precedente=bus;
      bus=bus->p_suivant;
    }
  }
  if (bus==NULL)
  {
    fprintf(fhistorique,"    allouer : pas de bloc assez grand !\n");
    return(-1);
  }

  /* Allocation de la zone */
  adresse=bus->bloc.adresse;
  if (bus->bloc.taille==t)
  {
    if (cel_precedente==NULL) *pp_liste_libre=bus->p_suivant;
    else cel_precedente->p_suivant=bus->p_suivant;
    free(bus);
    cpt_free++;
  }
  else
  {
    bus->bloc.taille-=t;
    bus->bloc.adresse+=t;
  }
  fprintf(fhistorique,"    allouer : adresse=%d\n",adresse);
  return(adresse);
}

int liberer(int adresse_libre, int taille_libre, struct liste_libre **pp_liste_libre)
/* Libere la zone memoire de taille taille_libre a l'adresse adresse_libre
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_libre *cel_precedente=NULL, *cel_suivante=NULL, *bus=*pp_liste_libre, *aux;

  fprintf(fhistorique,"    liberer %d %d\n",adresse_libre,taille_libre);
  if ((adresse_libre<0) || (adresse_libre+taille_libre>taille_memoire))
  {
    fprintf(fhistorique,"    liberer : intervalle invalide ?!\n");
    return(-1);
  }

  /* Recherche des cellules entourant la zone a liberer */
  while ((bus!=NULL) && (bus->bloc.adresse<adresse_libre))
  {
    cel_precedente=bus;
    bus=bus->p_suivant;
  }
  cel_suivante=bus;

  if ((cel_precedente!=NULL) && (adresse_libre<cel_precedente->bloc.adresse+cel_precedente->bloc.taille))
  {
    fprintf(fhistorique,"    liberer : zone deja libre ?!\n");
    return(-1);
  }
  if ((cel_suivante!=NULL) && (adresse_libre+taille_libre>cel_suivante->bloc.adresse))
  {
    fprintf(fhistorique,"    liberer : zone deja libre ?!\n");
    return(-1);
  }

  if ((cel_precedente!=NULL) && (cel_suivante!=NULL) && (cel_suivante->bloc.adresse==cel_precedente->bloc.adresse+cel_precedente->bloc.taille+taille_libre))
  {
    cel_precedente->p_suivant=cel_suivante->p_suivant;
    cel_precedente->bloc.taille+=taille_libre+cel_suivante->bloc.taille;
    free(cel_suivante);
    cpt_free++;
    fprintf(fhistorique,"    liberer : fusion de 3 zones\n");
    return(0);
  }

  if ((cel_precedente!=NULL) && (adresse_libre==cel_precedente->bloc.adresse+cel_precedente->bloc.taille))
  {
    cel_precedente->bloc.taille+=taille_libre;
    fprintf(fhistorique,"    liberer : fusion avec la cellule precedente\n");
    return(0);
  }

  if ((cel_suivante!=NULL) && (cel_suivante->bloc.adresse==adresse_libre+taille_libre))
  {
    cel_suivante->bloc.adresse-=taille_libre;
    cel_suivante->bloc.taille+=taille_libre;
    fprintf(fhistorique,"    liberer : fusion avec la cellule suivante\n");
    return(0);
  }

  aux=malloc(sizeof(struct liste_libre));
  if (aux==NULL)
  {
    fprintf(fhistorique,"    liberer : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  aux->bloc.adresse=adresse_libre;
  aux->bloc.taille=taille_libre;
  aux->p_suivant=cel_suivante;
  if (cel_precedente==NULL) *pp_liste_libre=aux;
  else cel_precedente->p_suivant=aux;
  fprintf(fhistorique,"    liberer : nouvelle cellule inseree\n");
  return(0);
}

int allouer_tas(char identite_tas[], int taille_tas, struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Ajoute un tas de nom identite_tas et de taille taille_tas
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_tas *aux;
  int i;

  fprintf(fhistorique,"  allouer_tas %s %d\n",identite_tas,taille_tas);
  if (existe_tas(identite_tas,*pp_liste_tas)!=NULL)
  {
    fprintf(fhistorique,"  allouer_tas : tas deja existant !\n");
    return(-1);
  }
  aux=malloc(sizeof(struct liste_tas));
  if (aux==NULL)
  {
    fprintf(fhistorique,"  allouer_tas : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  aux->tas.taille=taille_tas;
  strcpy(aux->tas.ident,identite_tas);
  aux->tas.compteur_references=0;
  for (i=0;i<MAX_REF;i++) aux->tas.liste_ident[i][0]='\0';
  if ((aux->tas.adresse=allouer(taille_tas,pp_liste_libre,pp_liste_tas))==-1)
  {
    free(aux);
    cpt_free++;
    fprintf(fhistorique,"  allouer_tas : erreur de allouer !\n");
    return(-1);
  }
  aux->p_suivant=*pp_liste_tas;
  *pp_liste_tas=aux;
  return(aux->tas.adresse);
}

int liberer_tas(char identite_tas[], struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Libere le tas de nom identite_tas
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct liste_tas *bus=*pp_liste_tas,*reference,*precedent=NULL;
  int i;

  fprintf(fhistorique,"  liberer_tas %s\n",identite_tas);
  while ((bus!=NULL) && (strcmp(identite_tas,bus->tas.ident)!=0))
  {
    precedent=bus;
    bus=bus->p_suivant;
  }
  if (bus==NULL)
  {
    fprintf(fhistorique,"  liberer_tas : tas inexistant !\n");
    return(-1);
  }
  if (bus->tas.compteur_references>0)
  {
    fprintf(fhistorique,"  liberer_tas : tas reference !\n");
    return(-1);
  }
  if (liberer(bus->tas.adresse,bus->tas.taille,pp_liste_libre)!=0)
  {
    fprintf(fhistorique,"  liberer_tas : erreur de liberer\n");
    return(-1);
  }

  for (i=0;i<MAX_REF;i++)
    if (bus->tas.liste_ident[i][0]!='\0')
    {
      reference=existe_tas(bus->tas.liste_ident[i],*pp_liste_tas);
      if (reference==NULL)
        fprintf(fhistorique,"  liberer_tas : tas reference a disparu ?!\n");
      else reference->tas.compteur_references--;
    }

  if (precedent==NULL) *pp_liste_tas=bus->p_suivant;
  else precedent->p_suivant=bus->p_suivant;
  free(bus);
  cpt_free++;
  return(0);
}

int ajouter_programme(char ident[], struct liste_tas *p_liste_tas, struct programme **pp_liste_programmes)
/* Ajoute le programme de nom ident
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct programme *aux,*bus_prog=*pp_liste_programmes;
  struct liste_tas *p_tas;

  fprintf(fhistorique,"  ajouter_programme %s\n",ident);
  p_tas=existe_tas(ident,p_liste_tas);
  if (p_tas==NULL)
  {
    fprintf(fhistorique,"  ajouter_programme : tas inexistant !\n");
    return(-1);
  }
  while (bus_prog!=NULL)
  {
    if (strcmp(ident,bus_prog->ident)==0)
    {
      fprintf(fhistorique,"  ajouter_programme : programme deja existant !\n");
      return(-1);
    }
    bus_prog=bus_prog->p_suivant;
  }
  aux=malloc(sizeof(struct programme));
  if (aux==NULL)
  {
    fprintf(fhistorique,"  ajouter_programme : erreur de malloc !\n");
    return(-1);
  }
  cpt_malloc++;
  strcpy(aux->ident,ident);
  aux->p_suivant=*pp_liste_programmes;
  p_tas->tas.compteur_references++;
  *pp_liste_programmes=aux;
  return(0);
}

int supprimer_programme(char ident[], struct liste_tas *p_liste_tas, struct programme **pp_liste_programmes)
/* Supprime le programme de nom ident
   Retourne -1 en cas d'erreur, 0 sinon */
{
  struct programme *precedent=NULL,*bus_prog=*pp_liste_programmes;
  struct liste_tas *p_tas;

  fprintf(fhistorique,"  supprimer_programme %s\n",ident);
  while ((bus_prog!=NULL) && (strcmp(ident,bus_prog->ident)!=0))
  {
    precedent=bus_prog;
    bus_prog=bus_prog->p_suivant;
  }
  if (bus_prog==NULL)
  {
    fprintf(fhistorique,"  supprimer_programme : programme inexistant !\n");
    return(-1);
  }
  if (precedent==NULL) *pp_liste_programmes=bus_prog->p_suivant;
  else precedent->p_suivant=bus_prog->p_suivant;
  free(bus_prog);
  cpt_free++;

  p_tas=existe_tas(ident,p_liste_tas);
  if (p_tas==NULL)
  {
    fprintf(fhistorique,"  supprimer_programme : le tas a disparu !?\n");
    return(-1);
  }
  p_tas->tas.compteur_references--;
  return(0);
}

int ajouter_lien(char tas_referencant[], char tas_reference[], struct liste_tas *p_liste_tas)
/* Ajoute un lien de tas_referencant vers tas_reference
   Retourne -1 en cas d'echec, 0 sinon */
{
  struct liste_tas *referencant,*reference;
  int i=0;

  fprintf(fhistorique,"  ajouter_lien %s %s\n",tas_referencant,tas_reference);
  referencant=existe_tas(tas_referencant,p_liste_tas);
  if (referencant==NULL)
  {
    fprintf(fhistorique,"  ajouter_lien : tas_referencant inexistant !\n");
    return(-1);
  }
  reference=existe_tas(tas_reference,p_liste_tas);
  if (reference==NULL)
  {
    fprintf(fhistorique,"  ajouter_lien : tas_reference inexistant !\n");
    return(-1);
  }

  while ((i<MAX_REF) && (referencant->tas.liste_ident[i][0]!='\0')) i++;
  if (i==MAX_REF)
  {
    fprintf(fhistorique,"  ajouter_lien : maximum de liens atteint !\n");
    return(-1);
  }
  strcpy(referencant->tas.liste_ident[i],tas_reference);
  reference->tas.compteur_references++;
  return(0);
}

int supprimer_lien(char tas_referencant[], char tas_reference[], struct liste_tas *p_liste_tas)
/* Supprime le lien de tas_referencant vers tas_reference
   Retourne -1 en cas d'echec, 0 sinon */
{
  struct liste_tas *referencant,*reference;
  int i=0;

  fprintf(fhistorique,"  supprimer_lien %s %s\n",tas_referencant,tas_reference);
  referencant=existe_tas(tas_referencant,p_liste_tas);
  if (referencant==NULL)
  {
    fprintf(fhistorique,"  supprimer_lien : tas_referencant inexistant !\n");
    return(-1);
  }
  reference=existe_tas(tas_reference,p_liste_tas);
  if (reference==NULL)
  {
    fprintf(fhistorique,"  supprimer_lien : tas_reference inexistant !\n");
    return(-1);
  }

  while ((i<MAX_REF) && (strcmp(referencant->tas.liste_ident[i],tas_reference))) i++;
  if (i==MAX_REF)
  {
    fprintf(fhistorique,"  supprimer_lien : lien inexistant !\n");
    return(-1);
  }
  referencant->tas.liste_ident[i][0]='\0';
  reference->tas.compteur_references--;
  return(0);
}

int ramasse_miettes_compteurs(struct liste_libre **pp_liste_libre, struct liste_tas **pp_liste_tas)
/* Libere les tas dont le compteur de reference est a zero
   Retoure -1 si aucun tas n'a ete libere, 0 sinon */
{
  struct liste_tas *bus,*aux;
  int n=0,compteur_miettes;

  fprintf(fhistorique,"  ramasse_miettes_compteurs\n");
  do
  {
    compteur_miettes=0;
    bus=*pp_liste_tas;
    while (bus!=NULL)
    {
      aux=bus->p_suivant;
      if (bus->tas.compteur_references==0)
        if (liberer_tas(bus->tas.ident,pp_liste_libre,pp_liste_tas)==0) compteur_miettes++;
      bus=aux;
    }
    n+=compteur_miettes;
  } while (compteur_miettes>0);
  fprintf(fhistorique,"  ramasse_miettes_compteurs : %d tas ramasse(s)\n",n);
  if (n==0) return(-1); else return(0);
}
